import numpy as np
import pandas as pd

try:
    import torch
except:
    print('Could not load PyTorch. Consider installing or fixing it for substantial speed gains. Using Numpy instead. ')

def get_comparable_pairs_torch(time, event):
    y = torch.from_numpy(time)
    e = torch.from_numpy(event)

    grid_x, grid_y = torch.meshgrid(y, y)
    grid_x = grid_x.tril()
    grid_y = grid_y.tril()
    diff_truth = grid_x - grid_y

    grid_ex, grid_ey = torch.meshgrid(e, e)
    valid_pairs =((diff_truth < 0) & (grid_ex == 1))
    res1 = torch.stack(torch.where(valid_pairs)).T

    diff_truth = grid_y - grid_x
    valid_pairs =((diff_truth < 0) & (grid_ey == 1))
    res2 = torch.stack(torch.where(valid_pairs)).T.flip(1)

    pairs = torch.cat([res1,res2]).numpy()
    return pairs

def cindex_by_pair_torch(v_high, v_low):
    eval_comparable = (v_high < v_low).float()
    eval_non_comparable = (v_high == v_low).float()
    return (eval_comparable+(eval_non_comparable*0.5))


def compute_univariate_score_of_torch(df, target_columns, biom_values):
    univ_scores = pd.DataFrame()
    for target in target_columns:
        # get comparable train pairs values
        pairs = get_comparable_pairs_torch(df[target_columns[target][0]].values,
                                           df[target_columns[target][1]].astype('bool').values)
        biom_values_low  = biom_values[pairs[:, 0]]
        biom_values_high = biom_values[pairs[:, 1]]

        # compute feature's signs
        features_cindex_by_pair = cindex_by_pair_torch(biom_values_high, biom_values_low)
        features_cindex = features_cindex_by_pair.mean(dim=0)
        univ_scores[target] = features_cindex.cpu().numpy()
    return univ_scores


def compute_univariate_score_torch(df, candidates, target_columns, device):
    max_chunk_size = 1e4
    if len(candidates) > max_chunk_size:
        n_chunks = len(candidates) / max_chunk_size
        chunks = np.array_split(candidates, n_chunks)
    else:
        chunks = [candidates]

    all_univ_scores = []
    for chunk in chunks: #tqdm(chunks, bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}'):
        biom_values = torch.as_tensor(df[chunk].values, device=device, dtype=torch.float)
        univ_scores = compute_univariate_score_of_torch(df, target_columns, biom_values)
        torch.cuda.empty_cache()
        univ_scores.index = chunk
        all_univ_scores.append(univ_scores)
    return pd.concat(all_univ_scores)


def compute_pvals_target_torch(scores, random_scores, device):
    scores = abs(scores-0.5)
    random_scores = torch.as_tensor(random_scores, device=device, dtype=torch.float)
    random_scores = abs(random_scores-0.5)
    n_random = len(random_scores)

    max_chunk_size = 1e2
    if len(scores) > max_chunk_size:
        n_chunks = len(scores) / max_chunk_size
        chunks = np.array_split(scores, n_chunks)
    else:
        chunks = [scores]

    all_pvals = []
    for chunk in chunks:
        chunk_values = torch.as_tensor(chunk, device=device, dtype=torch.float)
        chunk_values = chunk_values[:, None]
        pvals = ((random_scores >= chunk_values).sum(axis=1) + 1) / (n_random+1)
        all_pvals.append(pvals)
    all_pvals = torch.cat(all_pvals)
    return all_pvals.cpu().numpy()


def compute_pvals_torch(scores, random_scores, device):
    scores = scores.copy()
    for target in scores.columns:
        scores[target+'_pval'] = compute_pvals_target_torch(scores[target].values, random_scores[target].values, device)
        torch.cuda.empty_cache()
    return scores


import numpy as np
import pandas as pd
import ray
from lifelines import CoxPHFitter
from lifelines.statistics import proportional_hazard_test
from statsmodels.stats.outliers_influence import variance_inflation_factor
import scipy.spatial.distance as ssd
from scipy.cluster.hierarchy import linkage, leaves_list, optimal_leaf_ordering


def get_vif(df, cov):
    """
    Compute the variance inflation factor (VIF) for a given set of columns.
    """

    sdf = df[cov].dropna(how='any', axis=0)
    vif_data = {
        "feature": sdf.columns,
        "VIF": [variance_inflation_factor(sdf.values, i) for i in range(sdf.shape[1])]
    }
    max_vif = max(vif_data["VIF"])
    return {'vif': max_vif, 'candidate': cov[-1]}


@ray.remote
def get_all_vif_worker(df, candidates, confounders):
    """
    Compute VIF for multiple candidates.
    """

    import warnings
    warnings.filterwarnings("ignore")
    return [get_vif(df, confounders + [candidate]) for candidate in candidates]


def get_all_vif(df, candidates, confounders, n_workers):
    """
    Divide the task among workers and get VIF values.
    """

    candidates_chunks = np.array_split(candidates, n_workers)
    workers = [get_all_vif_worker.remote(ray.put(df), chunk, confounders) for chunk in candidates_chunks]
    all_res = sum(ray.get(workers), [])
    return pd.DataFrame(all_res).set_index('candidate')


def compute_cox_coef(df, candidate, targets, confounders, strata):
    """
    Compute univariate and multivariate cox coefficients (hazard ratios) for the given candidate.
    """
    result = pd.DataFrame(index=[candidate])

    for target, (duration_col, event_col) in targets.items():
        cph = CoxPHFitter()

        # Univariate analysis
        uni_vars = [candidate, duration_col, event_col]
        cph.fit(df[uni_vars], duration_col=duration_col, event_col=event_col)
        coef_uni = cph.summary.loc[candidate, 'coef']
        ph_test_uni = proportional_hazard_test(cph, df[uni_vars]).summary['p'].min()

        # Multivariate analysis
        multi_vars = uni_vars + confounders
        cph.fit(df[multi_vars], duration_col=duration_col, event_col=event_col, strata=strata, robust=True)
        coef_multi = cph.summary.loc[candidate, 'coef']
        ph_test_multi = proportional_hazard_test(cph, df[multi_vars]).summary['p'].min()

        # Store results
        result[target + '_uni'] = coef_uni
        result[target + '_multi'] = coef_multi
        result[target + '_uni_HRnorm'] = np.exp(abs(coef_uni))
        result[target + '_multi_HRnorm'] = np.exp(abs(coef_multi))
        result[target + '_ph_assumpt_uni'] = ph_test_uni
        result[target + '_ph_assumpt_multi'] = ph_test_multi
        result[target + '_diff_HRnorm'] = result[target + '_multi_HRnorm'] - result[target + '_uni_HRnorm']
        result[target + '_diffPerct_HRnorm'] = (result[target + '_diff_HRnorm'] / result[
            target + '_multi_HRnorm']) * 100

    return result


@ray.remote
def process_candidates(df, candidates, targets, confounders, strata):
    """
    Process a chunk of candidates using the compute_cox_coef function.
    """
    results = []
    for candidate in candidates:
        try:
            results.append(compute_cox_coef(df, candidate, targets, confounders, strata))
        except:
            pass
    return results


def get_all_cox_coef(df, candidates, targets, confounders, strata, n_workers):
    """
    Compute univariate and multivariate cox coefficients (hazard ratios) for the given candidate in parallel.
    Compute metrics to decide if a candidate is sensitive to confounders.
    """
    candidate_chunks = np.array_split(candidates, n_workers)
    df_id = ray.put(df)

    workers = [process_candidates.remote(df_id, chunk, targets, confounders, strata) for chunk in candidate_chunks]
    results = sum(ray.get(workers), [])

    return pd.concat(results)


def drop_high_corr(candidates_correlations, max_corr):
    """
    Randomly drop candidates that are too correlated to other candidates
    """

    upper = np.tril(np.full(candidates_correlations.shape, np.nan), k=0) + candidates_correlations
    nb_corrwith = (upper >= max_corr).sum()
    return nb_corrwith[nb_corrwith == 0].index.tolist()


def get_clusters(selected, candidates_correlations, max_corr):
    """
    Cluster selected items based on their correlations.
    """

    # If there is 0 or 1 items, return them as a single cluster
    if len(selected) < 2:
        return [selected]

    if max_corr == 1:
        return [[x] for x in selected]

    # Calculate distances from correlations and convert to a format suitable for linkage
    distances = 1 - candidates_correlations.loc[selected, selected]
    dist_array = ssd.squareform(distances)

    # Perform hierarchical clustering and get the optimal leaf order
    Z = linkage(dist_array, 'ward')
    order = leaves_list(optimal_leaf_ordering(Z, dist_array))

    # Order the selected items based on the clustering
    ordered_sel = np.array(selected)[order].tolist()

    # Create clusters based on the correlation threshold
    clusters = []
    current_cluster = [ordered_sel[0]]

    for fid in range(1, len(ordered_sel)):
        corr = candidates_correlations.loc[ordered_sel[fid], ordered_sel[fid - 1]]
        if corr > max_corr:
            current_cluster.append(ordered_sel[fid])
        else:
            clusters.append(current_cluster)
            current_cluster = [ordered_sel[fid]]

    clusters.append(current_cluster)
    return clusters


def primary_selection(df, candidates, confounders, strata, confound_check_corr, targets, max_corr, verbose,
                      n_workers=1):
    """Primary feature selection based on multiple criteria."""
    if verbose:
        print('\nStarting pre-selection')

    # Drop candidates with missing values
    candidates = df[candidates].dropna(axis=1).columns.tolist()
    if verbose:
        print(f"{len(candidates)} candidates remain after dropping candidates with NaNs.")

    # Drop candidates with constant values
    candidates = df[candidates].nunique().index[df[candidates].nunique() > 1].tolist()
    if verbose:
        print(f"{len(candidates)} candidates remain after dropping constant candidates.")

    # Drop candidates with low variance
    candidates = df[candidates].var()[df[candidates].var() > 1e-3].index.tolist()
    if verbose:
        print(f"{len(candidates)} candidates remain after dropping candidates with low variance.")

    if len(confound_check_corr) > 0:
        # Compute correlation between features and known confounders
        correlation_with_confounders = df[candidates].corrwith(df[confound_check_corr].max(axis=1),
                                                               method='spearman').abs()
        candidates = correlation_with_confounders[correlation_with_confounders < 0.5].index.tolist()
        if verbose:
            print(f"{len(candidates)} candidates remain after dropping candidates correlated with confounders.")

    if len(confound_check_corr) > 0:
        # Check multicollinearity with confounders
        all_vif = get_all_vif(df, candidates, confound_check_corr, n_workers)
        candidates = all_vif[all_vif['vif'] < 5].index.tolist()
        if verbose:
            print(
                f"{len(candidates)} candidates remain after dropping candidates with high multicollinearity with confounders.")

    if len(confounders) > 0:
        # Check for confoundings factor
        multivariate_eval = get_all_cox_coef(df, candidates, targets, confounders, strata, n_workers)
        candidates = multivariate_eval[
            multivariate_eval[[x for x in multivariate_eval.columns if '_diffPerct_HRnorm' in x]].abs().max(
                axis=1) < 10].index.tolist()
        if verbose:
            print(f'{len(candidates)} candidates remain after dropping candidates sensitive to confounders.')

    # Drop duplicated candidates
    candidates_correlations = df[candidates].corr(method='spearman').abs()
    candidates = drop_high_corr(candidates_correlations, max_corr=0.99)
    if verbose:
        print(f'{len(candidates)} candidates remain after dropping duplicated candidates.')

    # Group candidates into clusters of similar information based on their correlations
    corr_clusters = get_clusters(candidates, candidates_correlations, max_corr)

    return candidates, corr_clusters, candidates_correlations



import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import ray
from multipy.fdr import tst
from tqdm import tqdm
import seaborn as sns
from robi.evaluation import univariate_evaluation, score_of_random
from robi.preselection import primary_selection
from robi.utils import *

sns.set_theme(style="whitegrid")

def get_best_candidates_per_cluster(scores, clusters):
    representative = []
    for cluster in clusters:
        if len(cluster) > 1:
            max_pvals = scores.loc[cluster]
            best_idx = np.argmin(max_pvals.values)
            representative.append(max_pvals.index[best_idx])
        else:
            representative.append(cluster[0])
    return representative


def fdr_control_on_target(scores, clusters, permissiveness, target):
    candidates = get_best_candidates_per_cluster(scores[target + '_pval'], clusters)
    target_scores = scores.loc[candidates]
    target_scores['pass_' + target] = tst(target_scores[target + '_pval'].values, q=permissiveness)
    return target_scores


def get_permuted_scores(df, candidates, scores_random, targets, device, n_permut, verbose):
    cols_to_permut = sum([list(targets[x]) for x in targets], [])
    all_pscores = []
    for it in tqdm(range(n_permut),
                   desc='Computing scores of permutations',
                   bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',
                   disable=not verbose):
        dfp = df.copy()
        dfp.loc[dfp.index, cols_to_permut] = dfp.loc[np.random.permutation(dfp.index), cols_to_permut].values
        pscores = univariate_evaluation(dfp, candidates, targets, device, scores_random)
        all_pscores.append(pscores)
    return all_pscores


@ray.remote
def get_nfp_all_permissiveness_worker(pscores, chunk, corr_clusters, permissivenesses, target):
    all_res = []
    for x in chunk:
        s = pscores[x]
        sel_candidates = get_best_candidates_per_cluster(s[target + '_pval'], corr_clusters)
        target_scores = s.loc[sel_candidates]
        all_pass = {}
        for p in permissivenesses:
            all_res.append({
                'permissiveness': p,
                'n_fp_' + target: tst(target_scores[target + '_pval'].values, q=p).sum()
            })
    return all_res


def get_nfp_all_permissiveness(perm_values, target, pscores, corr_clusters, n_workers):
    n_permut = len(pscores)
    chunks = np.array_split(np.arange(n_permut), n_workers)
    pscores_id = ray.put(pscores)
    corr_clusters_id = ray.put(corr_clusters)
    workers = [get_nfp_all_permissiveness_worker.remote(pscores_id, x,
                                                        corr_clusters_id, perm_values, target) for x in chunks]
    res = sum(ray.get(workers), [])
    res = pd.DataFrame(res)
    res = res.set_index('permissiveness')
    return res


def fdr_control(scores, clusters, permissiveness, targets):
    all_target_scores = {}
    for target in targets:
        all_target_scores[target] = fdr_control_on_target(scores, clusters, permissiveness, target)
    return all_target_scores


def get_nsel_by_permissiveness(scores, corr_clusters):
    all_n_sel = []
    targets = np.unique([x.split('_pval')[0] for x in scores.columns]).tolist()
    for permissiveness in np.arange(0.01, 1.01, 0.01).round(2):
        sel = fdr_control(scores, corr_clusters, permissiveness, targets=targets)
        r = {'permissiveness': permissiveness}
        for target in targets:
            r['sel_' + target] = sel[target][sel[target]['pass_' + target]].index.tolist()
            r['n_sel_' + target] = sel[target]['pass_' + target].sum()
        all_n_sel.append(r)
    return pd.DataFrame(all_n_sel).set_index('permissiveness')


def plot_permissiveness_effect(scores, pscores, corr_clusters, targets, n_workers, verbose):
    nsel_by_permissiveness = get_nsel_by_permissiveness(scores, corr_clusters)
    plotid = 0
    if verbose:
        plt.figure(figsize=(12, 5), layout='constrained')
    nfp_all_permissiveness_by_target = {}
    for target in targets:
        if verbose:
            plotid += 1
            plt.subplot(1, len(list(targets.keys())), plotid)
        nfp_all_permissiveness = get_nfp_all_permissiveness(nsel_by_permissiveness.index, target,
                                                            pscores, corr_clusters, n_workers)
        nfp_all_permissiveness_by_target[target] = nfp_all_permissiveness
        if verbose:
            sns.lineplot(nsel_by_permissiveness, x='permissiveness', y='n_sel_' + target,
                         label='Number of\nselected candidates', color='#ec6602')
            sns.lineplot(nfp_all_permissiveness, x='permissiveness', y='n_fp_' + target,
                         label='Number of\nfalse positives', errorbar=('pi', 95), color='#009999')
            plt.title(target)
            plt.xlabel('Requested False Discovery Rate (TST\'s FDR)')
            plt.ylabel('')
            if plotid == 1:
                plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
            else:
                plt.legend('', frameon=False)
    if verbose:
        plt.show()
    return nsel_by_permissiveness, nfp_all_permissiveness_by_target


def get_selection_results(n_fp, targets, nsel_by_permissiveness, nfp_all_permissiveness):
    all_rows = []
    for target in targets:
        row = {}
        row['target'] = target
        mean_nfp = nfp_all_permissiveness[target].groupby('permissiveness').mean()
        row['permissiveness'] = mean_nfp[mean_nfp['n_fp_' + target] < n_fp].index[-1]

        row['selection'] = nsel_by_permissiveness.loc[row['permissiveness'], 'sel_' + target]
        row['n_selected'] = nsel_by_permissiveness.loc[row['permissiveness'], 'n_sel_' + target]

        sel_fp = nfp_all_permissiveness[target].loc[row['permissiveness'], 'n_fp_' + target].values
        row['prob_only_FP'] = ((sel_fp >= row['n_selected']).sum() + 1) / (len(sel_fp) + 1)
        row['n_FP_(CI)'] = '%.2f' % np.mean(sel_fp) + ' (%.2f' % np.percentile(sel_fp, 2.5) + '-%.2f)' % np.percentile(
            sel_fp, 97.5)
        all_rows.append(row)
    all_rows = pd.DataFrame(all_rows)
    all_rows = all_rows.set_index('target')
    return all_rows.T


def make_selection(df,
                   candidates,
                   targets,
                   confounders=[],
                   strata=None,
                   confounders_to_norm=[],
                   confound_check_corr=[],
                   max_corr=0.5,
                   n_workers=1,
                   n_random=100,
                   n_permut_nfp=10,
                   verbose=True,
                   device="cpu"):
    if verbose:
        print('Selection started...')

    if torch_installed(verbose=True) and device != 'cuda':
        print(f"device is {device}. If you have a GPU, switching to cuda will provide significant speed gains.")

    # drop cases where confounders are not defined
    df = df[df[confounders].isna().sum(axis=1) == 0]

    df = normalize_columns(df, candidates + confounders_to_norm)
    if verbose:
        check_confounders(df, confounders, targets, strata)

    # perform the preselection and
    sel_candidates, corr_clusters, candidates_correlations = primary_selection(df,
                                                                               candidates,
                                                                               confounders,
                                                                               strata,
                                                                               confound_check_corr,
                                                                               targets,
                                                                               max_corr,
                                                                               verbose,
                                                                               n_workers)

    scores_random = score_of_random(df, targets, device, n_random)
    scores = univariate_evaluation(df, sel_candidates, targets, device, scores_random)
    if verbose:
        check_n_random(scores, targets, n_random)

    pscores = get_permuted_scores(df, sel_candidates, scores_random, targets, device, n_permut_nfp, verbose)

    nsel_by_permissiveness, nfp_all_permissiveness = plot_permissiveness_effect(scores, pscores,
                                                                                corr_clusters, targets,
                                                                                n_workers, verbose)

    for target in targets:
        mean_nfp = nfp_all_permissiveness[target].groupby('permissiveness').mean()
        ci_nfp_low = nfp_all_permissiveness[target].groupby('permissiveness').quantile(0.025).round(2).astype('str')
        ci_nfp_high = nfp_all_permissiveness[target].groupby('permissiveness').quantile(0.975).round(2).astype('str')
        mean_nfp = mean_nfp.round(1).astype('str') + ' (' + ci_nfp_low + '-' + ci_nfp_high + ')'
        nsel_by_permissiveness = pd.concat([nsel_by_permissiveness, mean_nfp], axis=1)

    return nsel_by_permissiveness, scores



from lifelines import CoxPHFitter
from lifelines.statistics import proportional_hazard_test


def check_n_random(scores, targets, n_random):
    """
    Check if the number of random features is sufficient based on the minimum p-value.
    The number of random features that are better than the candidates should be high enough
    to ensure precise p-value estimation, event for candidates with really high score.
    Prints a warning if the number of random features is considered insufficient.
    """

    # Calculate the minimum p-value across all targets
    min_pval = scores[[f"{t}_pval" for t in targets]].min().min()
    n_sup = int(min_pval * n_random)

    # Check if the number of superior random features is less than 200
    if n_sup < 200:
        if n_sup > 1:
            recommended_n_random = int(200 / n_sup * n_random)
            print(f"Warning: only {n_sup} random features were better than the best candidate."
                  f"\nConsider increasing n_random. We recommend at least {recommended_n_random:.2E} random features.")
        else:
            print('Warning: no random features were better than the best candidate. Consider increasing n_random.')
    else:
        print('n_random is high enough')


def normalize_columns(df, columns):
    """
    Normalize (z-score) specified columns in the dataframe.
    """

    df.loc[:, columns] = (df[columns] - df[columns].mean()) / df[columns].std()
    return df


def check_proportional_hazard_for_target(df, target, targets, confounders, strata):
    """
    Check proportional hazard assumption for a specific target.
    """

    # Construct column names based on target
    duration_col = targets[target][0]
    event_col = targets[target][1]

    # Filter necessary columns from dataframe
    df_filtered = df[confounders + [duration_col, event_col]]

    # Fit the Cox Proportional Hazard model
    cph = CoxPHFitter()
    cph.fit(df_filtered, duration_col=duration_col, event_col=event_col, strata=strata, robust=True)

    # Test for proportional hazard assumption
    results = proportional_hazard_test(cph, df_filtered)

    # Check if any confounder violates the assumption
    for confounder, row in results.summary.iterrows():
        if row['p'] < 0.05:
            print(f"{confounder} violates proportional hazard assumptions for {target} (p<{row['p']:.2f}).",
                  "Consider adding to the strata list.")
            return False

    return True


def check_proportional_hazard(df, confounders, targets, strata):
    """
    Verify the proportional hazard assumption for specified targets and confounders.
    Prints the results of the verification.
    """

    all_tests_ok = all(
        check_proportional_hazard_for_target(df, target, targets, confounders, strata) for target in targets)

    if all_tests_ok:
        print('Confounders follow proportional hazard assumptions')


def check_confounders(df, confounders, targets, strata):
    """
    Check if the number of confounders exceeds the allowed limit for each target and
    verify if the confounders follows the proportional hazard assumption.
    """

    print('\nVerifying confounders settings...')

    # Calculate the maximum allowed number of confounders for each target
    # Use the rule-of-thumb stating that you should have at least
    for target, (_, event_col) in targets.items():
        n_event = df[event_col].sum()
        max_n_conf = int(n_event / 8 + 0.5)

        if len(confounders) > max_n_conf:
            print(f'Too many confounders for {target}. Max recommended: {max_n_conf}')

    # Check for proportional hazards (function not provided in the original code)
    check_proportional_hazard(df, confounders, targets, strata)


def torch_installed(verbose=False):
    try:
        import torch
        return True
    except:
        if verbose:
            print('Could not load PyTorch. Consider installing or fixing it for substantial speed gains. Using Numpy instead. ')
        return False


import numpy as np
import pandas as pd
from robi.evaluation_npy import *
from robi.evaluation_torch import *
from robi.utils import torch_installed

def compute_pvals(scores, scores_random, device):
    if torch_installed():
        return compute_pvals_torch(scores, scores_random, device)
    return compute_pvals_npy(scores, scores_random)


def compute_univariate_score(df_random, random_cols, targets, device):
    if torch_installed():
        return compute_univariate_score_torch(df_random, random_cols, targets, device)
    return compute_univariate_score_npy(df_random, random_cols, targets)


def score_of_random(df, targets, device, n_random):
    df_random = pd.DataFrame(data=np.random.uniform(0,1, (df.shape[0], int(n_random))), index=df.index)
    random_cols = df_random.columns
    df_random = pd.concat([df[sum([list(targets[x]) for x in targets], [])], df_random], axis=1)
    return compute_univariate_score(df_random, random_cols, targets, device)


def univariate_evaluation(df, candidates, targets, device, scores_random):
    scores = compute_univariate_score(df, candidates, targets, device)
    return compute_pvals(scores, scores_random, device)